#!/usr/bin/python
# coding: utf8

from .what3words import Geocoder  # noqa: F401
from .what3words import Circle  # noqa: F401
from .what3words import Coordinates  # noqa: F401
from .what3words import BoundingBox  # noqa: F401

from .version import __version__ as v

__title__ = "what3words"
__author__ = "what3words"
__author_email__ = "development@what3words.com"
__version__ = v
__license__ = "MIT"
__copyright__ = "Copyright (c) 2024 what3words"
